from enum import Enum
from p1afempy.data_structures import CoordinatesType
import numpy as np
from dataclasses import dataclass


@dataclass
class WeightsAndIntegrationPoints:
    weights: np.ndarray
    integration_points: CoordinatesType


class CubatureRuleEnum(Enum):
    MIDPOINT = 1
    LAUFFER_LINEAR = 2
    SMPLX1 = 3


@dataclass
class CubatureRule:
    weights_and_integration_points: WeightsAndIntegrationPoints
    degree_of_exactness: int
    name: str



import numpy as np
from p1afempy.data_structures import CoordinatesType
from triangle_cubature.cubature_rule \
    import WeightsAndIntegrationPoints


def transform_weights_and_integration_points(
    weights_and_integration_points: WeightsAndIntegrationPoints,
    physical_triangle: CoordinatesType
) -> WeightsAndIntegrationPoints:
    reference_weights = weights_and_integration_points.weights
    reference_integration_points = \
        weights_and_integration_points.integration_points

    jacobian = get_jacobian(physical_triangle=physical_triangle)

    transformed_weights = transform_weights(
        reference_weights=reference_weights,
        jacobian=jacobian)
    transformed_integration_points = transform_integration_points(
        reference_integration_points=reference_integration_points,
        p1=physical_triangle[0, :],
        jacobian=jacobian)

    return WeightsAndIntegrationPoints(
        weights=transformed_weights,
        integration_points=transformed_integration_points)


def transform_weights(reference_weights: np.ndarray,
                      jacobian: np.ndarray) -> np.ndarray:
    jacobian_determinant = np.linalg.det(jacobian)
    return jacobian_determinant * reference_weights


def transform_integration_points(
        reference_integration_points: CoordinatesType,
        p1: np.ndarray,
        jacobian: np.ndarray) -> CoordinatesType:

    transformed_integration_points = p1 \
        + reference_integration_points.dot(jacobian.T)

    return np.array(transformed_integration_points)


def get_jacobian(physical_triangle: CoordinatesType) -> np.ndarray:
    p1 = physical_triangle[0, :]
    p2 = physical_triangle[1, :]
    p3 = physical_triangle[2, :]
    return np.column_stack((p2 - p1, p3 - p1))



from p1afempy.data_structures import \
    CoordinatesType, ElementsType, BoundaryConditionType
from triangle_cubature.cubature_rule \
    import WeightsAndIntegrationPoints
from triangle_cubature.transformations import \
    transform_weights_and_integration_points
import numpy as np


def integrate_on_triangle(
        f: BoundaryConditionType, triangle: CoordinatesType,
        weights_and_integration_points: WeightsAndIntegrationPoints) -> float:
    transformed = transform_weights_and_integration_points(
        weights_and_integration_points=weights_and_integration_points,
        physical_triangle=triangle)
    return np.dot(transformed.weights, f(transformed.integration_points))


def integrate_on_mesh(
        f: BoundaryConditionType,
        coordinates: CoordinatesType,
        elements: ElementsType,
        weights_and_integration_points: WeightsAndIntegrationPoints) -> float:
    sum = 0.
    for triangle in elements:
        sum += integrate_on_triangle(
            f=f,
            triangle=coordinates[triangle, :],
            weights_and_integration_points=weights_and_integration_points)
    return sum



import numpy as np
from triangle_cubature.cubature_rule \
    import CubatureRule, CubatureRuleEnum, WeightsAndIntegrationPoints


def get_rule(rule: CubatureRuleEnum) -> CubatureRule:
    """
    given a cubature rule, returns the corresponding
    weight(s) and integration point(s)

    Notes
    -----
    - the rules correspond to the rules as specified in [1]

    References
    ----------
    - [1] Stenger, Frank.
      'Approximate Calculation of Multiple Integrals (A. H. Stroud)'.
      SIAM Review 15, no. 1 (January 1973): 234-35.
      https://doi.org/10.1137/1015023. p. 306-315
    """
    if rule == CubatureRuleEnum.MIDPOINT:
        weights = np.array([1./2.])
        integration_points = np.array([1./3., 1./3.]).reshape(1, 2)
        weights_and_integration_points = WeightsAndIntegrationPoints(
            weights=weights,
            integration_points=integration_points)
        name = 'midpoint'
        degree_of_exactness = 1

        return CubatureRule(
            weights_and_integration_points=weights_and_integration_points,
            degree_of_exactness=degree_of_exactness,
            name=name)

    if rule == CubatureRuleEnum.LAUFFER_LINEAR:
        integration_points = np.array([
            [0., 0.],
            [1., 0.],
            [0., 1.]
        ])
        weights = np.array([
            1/3 * 0.5,
            1/3 * 0.5,
            1/3 * 0.5
        ])
        weights_and_integration_points = WeightsAndIntegrationPoints(
            weights=weights,
            integration_points=integration_points)
        name = 'lauffer-linear'
        degree_of_exactness = 1

        return CubatureRule(
            weights_and_integration_points=weights_and_integration_points,
            degree_of_exactness=degree_of_exactness,
            name=name)

    if rule == CubatureRuleEnum.SMPLX1:
        r = (1.)/(6.)
        s = 2./3.
        integration_points = np.array([
            [r, r],
            [r, s],
            [s, r]
        ])
        weights = np.array([
            1/3 * 0.5,
            1/3 * 0.5,
            1/3 * 0.5
        ])
        weights_and_integration_points = WeightsAndIntegrationPoints(
            weights=weights,
            integration_points=integration_points)
        name = 'SMPLX1'
        degree_of_exactness = 2

        return CubatureRule(
            weights_and_integration_points=weights_and_integration_points,
            degree_of_exactness=degree_of_exactness,
            name=name)

    raise ValueError('specified rule does not exist.')
